// src/lib.rs

use pyo3::prelude::*;
use pyo3::types::{IntoPyDict, PyAny, PyDict, PyList, PyType};
use pyo3::wrap_pyfunction;

use numpy::{PyArray1, PyArray2, IntoPyArray};
use ndarray::{Array1, Array2, ArrayView1, ArrayView2, s, Axis};
use statrs::distribution::{Continuous, Normal, ContinuousCDF};
use std::collections::HashMap;
use std::fmt::Write as _;  // for building cluster-keys

/// Add significance stars.
fn add_significance_stars(p: f64) -> &'static str {
    if p < 0.01 {
        "***"
    } else if p < 0.05 {
        "**"
    } else if p < 0.1 {
        "*"
    } else {
        ""
    }
}

/// A tiny helper to check if we have a Logit or Probit from statsmodels
fn detect_model_type(model: &PyAny) -> Result<bool, PyErr> {
    // Return is_logit: true if logit, false if probit
    let model_obj = model.getattr("model").unwrap_or(model);
    let cls: String = model_obj
        .getattr("__class__")?
        .getattr("__name__")?
        .extract()?;
    let lc = cls.to_lowercase();
    if lc == "logit" {
        Ok(true) // is_logit = true
    } else if lc == "probit" {
        Ok(false) // is_logit = false
    } else {
        Err(pyo3::exceptions::PyValueError::new_err(
            format!("ame: only Logit or Probit supported, got {cls}"),
        ))
    }
}

/// Downcast a Python object to NumPy PyArray2<f64> => ndarray::ArrayView2<f64>.
///
/// Marked unsafe because .as_array() in pyo3-numpy is unsafe, trusting Python memory is valid.
fn as_array2_f64<'py>(obj: &'py PyAny) -> PyResult<ArrayView2<'py, f64>> {
    let pyarray = obj.downcast::<PyArray2<f64>>()?;
    let view = unsafe { pyarray.as_array() };
    Ok(view)
}

/// Similarly, for 1D arrays.
fn as_array1_f64<'py>(obj: &'py PyAny) -> PyResult<ArrayView1<'py, f64>> {
    let pyarray = obj.downcast::<PyArray1<f64>>()?;
    let view = unsafe { pyarray.as_array() };
    Ok(view)
}

/// Evaluate logistic or normal cdf
fn cdf_logit_probit(is_logit: bool, z: f64) -> f64 {
    if is_logit {
        // logistic cdf
        1.0 / (1.0 + (-z).exp())
    } else {
        // normal cdf
        let dist = Normal::new(0.0, 1.0).unwrap();
        dist.cdf(z)
    }
}
/// Evaluate logistic or normal pdf
fn pdf_logit_probit(is_logit: bool, z: f64) -> f64 {
    if is_logit {
        // logistic pdf
        let e = z.exp();
        e / (1.0 + e).powi(2)
    } else {
        // normal pdf
        (-0.5 * z * z).exp() / (2.0 * std::f64::consts::PI).sqrt()
    }
}

/// Derivative of the pdf with respect to z
fn pdf_deriv_logit_probit(is_logit: bool, z: f64) -> f64 {
    if is_logit {
        // For logistic: f'(z) = f(z) * (1 - 2 * F(z))
        pdf_logit_probit(is_logit, z) * (1.0 - 2.0 * cdf_logit_probit(is_logit, z))
    } else {
        // For normal: f'(z) = -z * phi(z)
        -z * pdf_logit_probit(is_logit, z)
    }
}

/// The main function that calculates Logit/Probit AME in the same style as the original Probit code.
/// 
/// An extra parameter `se_method` (optional, default "rust") controls the SE calculation:
///  - "rust": uses your original gradient (Jacobian) calculation
///  - "sm": uses a statsmodels-style gradient for continuous variables
#[pyfunction]
fn ame<'py>(
    py: Python<'py>,
    model: &'py PyAny,              // Could be Logit or Probit
    chunk_size: Option<usize>,      // Optional chunk size
    se_method: Option<&str>,        // "rust" (default) or "sm"
) -> PyResult<&'py PyAny> {
    // Determine which SE method to use.
    let se_method_str = se_method.unwrap_or("rust");

    // 1) detect Logit vs Probit
    let is_logit = detect_model_type(model)?;

    // 2) read params (handle pandas Series)
    let params_obj = model.getattr("params")?;
    let params_pyarray = if let Ok(values) = params_obj.getattr("values") {
        values.downcast::<PyArray1<f64>>()?
    } else {
        params_obj.downcast::<PyArray1<f64>>()?
    };
    let beta = unsafe { params_pyarray.as_array() };

    // 3) read cov (handle pandas DataFrame)
    let cov_obj = model.call_method0("cov_params")?;
    let cov_pyarray = if let Ok(values) = cov_obj.getattr("values") {
        values.downcast::<PyArray2<f64>>()?
    } else {
        cov_obj.downcast::<PyArray2<f64>>()?
    };
    let cov_beta = unsafe { cov_pyarray.as_array() };

    // 4) Get model object and handle exog (X) and exog_names
    let model_obj = model.getattr("model").unwrap_or(model);
    let exog_py = model_obj.getattr("exog")?;
    let (x_pyarray, exog_names) = if let Ok(values) = exog_py.getattr("values") {
        (
            values.downcast::<PyArray2<f64>>()?,
            exog_py.getattr("columns")?.extract::<Vec<String>>()?
        )
    } else {
        (
            exog_py.downcast::<PyArray2<f64>>()?,
            model_obj.getattr("exog_names")?.extract::<Vec<String>>()?
        )
    };

    let X = unsafe { x_pyarray.as_array() };
    let (n, k) = (X.nrows(), X.ncols());
    let chunk = chunk_size.unwrap_or(n);

    // 5) Identify intercept columns
    let intercept_indices: Vec<usize> = exog_names
        .iter()
        .enumerate()
        .filter_map(|(i, nm)| {
            let ln = nm.to_lowercase();
            if ln == "const" || ln == "intercept" {
                Some(i)
            } else {
                None
            }
        })
        .collect();

    // 6) Identify discrete columns => strictly 0/1
    let is_discrete: Vec<usize> = exog_names
        .iter()
        .enumerate()
        .filter_map(|(j, _)| {
            if intercept_indices.contains(&j) {
                None
            } else {
                let col_j = X.column(j);
                if col_j.iter().all(|&v| v == 0.0 || v == 1.0) {
                    Some(j)
                } else {
                    None
                }
            }
        })
        .collect();

    // 7) Prepare accumulators.
    // We keep two sets of accumulators:
    //  - one for your "rust" (original) method,
    //  - one for the statsmodels ("sm") method.
    let mut sum_ame = vec![0.0; k];
    let mut partial_jl_sums = vec![0.0; k * k];
    let mut sm_sum_ame = vec![0.0; k];
    let mut sm_partial_jl_sums = vec![0.0; k * k];
    let normal = Normal::new(0.0, 1.0).unwrap();

    // 8) Process data in chunks
    let mut idx_start = 0;
    while idx_start < n {
        let idx_end = (idx_start + chunk).min(n);
        let x_chunk = X.slice(s![idx_start..idx_end, ..]);
        let z_chunk = x_chunk.dot(&beta);  // shape(n_chunk)
        let pdf_chunk = z_chunk.mapv(|z| pdf_logit_probit(is_logit, z));

        // Handle discrete variables (same for both methods)
        for &j in &is_discrete {
            let xj_col = x_chunk.column(j);
            let b_j = beta[j];
            // Compute z for x_j set to 1 and 0:
            let delta_j1 = (1.0 - &xj_col).mapv(|x| x * b_j);
            let delta_j0 = xj_col.mapv(|x| -x * b_j);
            let z_j1 = &z_chunk + &delta_j1;
            let z_j0 = &z_chunk + &delta_j0;

            let cdf_j1 = z_j1.mapv(|z| cdf_logit_probit(is_logit, z));
            let cdf_j0 = z_j0.mapv(|z| cdf_logit_probit(is_logit, z));
            let effect_sum = cdf_j1.sum() - cdf_j0.sum();
            sum_ame[j] += effect_sum;
            sm_sum_ame[j] += effect_sum;

            let pdf_j1 = z_j1.mapv(|z| pdf_logit_probit(is_logit, z));
            let pdf_j0 = z_j0.mapv(|z| pdf_logit_probit(is_logit, z));
            for l in 0..k {
                let grad = if l == j {
                    pdf_j1.sum()
                } else {
                    let x_l = x_chunk.column(l);
                    let diff_pdf = &pdf_j1 - &pdf_j0;
                    diff_pdf.dot(&x_l)
                };
                partial_jl_sums[j * k + l] += grad;
                sm_partial_jl_sums[j * k + l] += grad;
            }
        }

        // Handle continuous variables.
        for j in 0..k {
            if intercept_indices.contains(&j) || is_discrete.contains(&j) {
                continue;
            }
            let b_j = beta[j];
            if se_method_str == "rust" {
                // Original method: update using b_j * pdf
                sum_ame[j] += b_j * pdf_chunk.sum();
                for l in 0..k {
                    let grad = if j == l {
                        pdf_chunk.sum()
                    } else {
                        let x_l = x_chunk.column(l);
                        -b_j * (&z_chunk * &x_l).dot(&pdf_chunk)
                    };
                    partial_jl_sums[j * k + l] += grad;
                }
            } else if se_method_str == "sm" {
                // statsmodels-style: use the derivative of the pdf.
                let fprime_chunk = z_chunk.mapv(|z| pdf_deriv_logit_probit(is_logit, z));
                sm_sum_ame[j] += b_j * pdf_chunk.sum();
                for l in 0..k {
                    let x_l = x_chunk.column(l);
                    let term = (&x_l * &fprime_chunk).sum();
                    let grad = if j == l {
                        pdf_chunk.sum() + b_j * term
                    } else {
                        b_j * term
                    };
                    sm_partial_jl_sums[j * k + l] += grad;
                }
            }
        }

        idx_start = idx_end;
    }

    // 9) Average sums over n.
    let final_ame: Vec<f64> = if se_method_str == "sm" {
        sm_sum_ame.iter().map(|v| v / (n as f64)).collect()
    } else {
        sum_ame.iter().map(|v| v / (n as f64)).collect()
    };

    // Build gradient matrix: shape (k, k)
    let mut grad_ame = Array2::<f64>::zeros((k, k));
    for j in 0..k {
        for l in 0..k {
            let value = if se_method_str == "sm" {
                sm_partial_jl_sums[j * k + l] / (n as f64)
            } else {
                partial_jl_sums[j * k + l] / (n as f64)
            };
            grad_ame[[j, l]] = value;
        }
    }

    // cov => grad_ame * cov_beta * grad_ame^T
    let cov_ame = grad_ame.dot(&cov_beta).dot(&grad_ame.t());
    let var_ame: Vec<f64> = cov_ame.diag().iter().map(|&v| v.max(0.0)).collect();
    let se_ame: Vec<f64> = var_ame.iter().map(|&v| v.sqrt()).collect();

    // 10) Build final results
    let (mut dy_dx, mut se_err, mut z_vals, mut p_vals, mut sig) =
        (Vec::new(), Vec::new(), Vec::new(), Vec::new(), Vec::new());
    let (mut conf_low, mut conf_high) = (Vec::new(), Vec::new());
    let mut names_out = Vec::new();

    for j in 0..k {
        if intercept_indices.contains(&j) {
            continue;
        }
        let dy = final_ame[j];
        let s = se_ame[j];
        dy_dx.push(dy);
        se_err.push(s);
        if s > 1e-15 {
            let z = dy / s;
            let p = 2.0 * (1.0 - normal.cdf(z.abs()));
            z_vals.push(z);
            p_vals.push(p);
            sig.push(add_significance_stars(p));
            // Compute 95% confidence interval using 1.96 * SE
            conf_low.push(dy - 1.96 * s);
            conf_high.push(dy + 1.96 * s);
        } else {
            z_vals.push(0.0);
            p_vals.push(1.0);
            sig.push("");
            conf_low.push(dy);
            conf_high.push(dy);
        }
        names_out.push(exog_names[j].clone());
    }

    // 11) Create DataFrame with extra confidence interval columns.
    let pd = py.import("pandas")?;
    let data = PyDict::new(py);
    data.set_item("dy/dx", &dy_dx)?;
    data.set_item("Std. Err", &se_err)?;
    data.set_item("z", &z_vals)?;
    data.set_item("Pr(>|z|)", &p_vals)?;
    data.set_item("Conf. Int. Low", &conf_low)?;
    data.set_item("Conf. Int. Hi", &conf_high)?;
    data.set_item("Significance", &sig)?;

    let kwargs = PyDict::new(py);
    kwargs.set_item("data", data)?;
    kwargs.set_item("index", &names_out)?;

    let df = pd.call_method("DataFrame", (), Some(kwargs))?;
    Ok(df)
}

#[pymodule]
fn febolt(_py: Python, m: &PyModule) -> PyResult<()> {
    m.add_function(wrap_pyfunction!(ame, m)?)?;
    Ok(())
}

